"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var CollapsibleMenu = /** @class */ (function () {
    function CollapsibleMenu(rootElement, rootLIElements, stringService, heightService) {
        var _this = this;
        // Called after setLIElementFilteredState. setLIElementFilteredState sets some elements to "display: none" through the filter-hidden class.
        // It is necessary to process the entire tree before toggling heights, since the final height of each element must be known.
        this.setFilteredLIElementHeight = function (liElement, allParentsAlreadyExpanded) {
            var directChildULElement = liElement.querySelector('ul');
            // Leaves don't need their heights toggled
            if (directChildULElement) {
                var alreadyExpanded = liElement.classList.contains('expanded');
                var toBeExpanded = liElement.classList.contains('filter-expanded');
                var closestChildLIElements = directChildULElement.children;
                for (var i = 0; i < closestChildLIElements.length; i++) {
                    _this.setFilteredLIElementHeight(closestChildLIElements[i], allParentsAlreadyExpanded && alreadyExpanded);
                }
                if (toBeExpanded && !alreadyExpanded) {
                    if (allParentsAlreadyExpanded) {
                        _this._heightService.toggleHeightWithTransition(directChildULElement, liElement);
                    }
                    else {
                        _this._heightService.toggleHeightWithoutTransition(directChildULElement, liElement);
                    }
                }
                else if (!toBeExpanded && alreadyExpanded) {
                    _this._heightService.toggleHeightWithTransition(directChildULElement, liElement);
                }
            }
        };
        this._stringService = stringService;
        this._heightService = heightService;
        this._rootElement = rootElement;
        this._rootLIElements = rootLIElements;
    }
    CollapsibleMenu.prototype.filter = function (filterValue) {
        if (filterValue === void 0) { filterValue = null; }
        for (var i = 0; i < this._rootLIElements.length; i++) {
            var rootLIElement = this._rootLIElements[i];
            this.setLIElementFilteredState(rootLIElement, filterValue);
            this.setFilteredLIElementHeight(rootLIElement, true);
        }
    };
    // An LI element can have three possible states, filter-expanded, filter-match and filter-hidden. filter-hidden is mutually exclusive from the other
    // two states, also an element can have the filter-match state but not the filter-expanded state. An LI element has the filter-match state if its 
    // contents contain the filter value. It has the filter-expanded state if any of its children has the filter-match or filter-expanded states.
    // Lastly, it has the filter-hidden state if it does not have either of the filter-match or filter-expanded states.
    //
    // Since LI elements are organized as a tree and the states of children are required to determine the states of parents, this state setting function 
    // performs a depth first search in post-order.
    CollapsibleMenu.prototype.setLIElementFilteredState = function (liElement, filterValue) {
        if (filterValue === void 0) { filterValue = null; }
        // Reset
        this.resetLIElementState(liElement);
        var directChildULElement = liElement.querySelector('ul');
        var toBeExpanded = false;
        // Skip leaves
        if (directChildULElement) {
            var closestChildLIElements = directChildULElement.children;
            for (var i = 0; i < closestChildLIElements.length; i++) {
                var childLIElement = closestChildLIElements[i];
                this.setLIElementFilteredState(childLIElement, filterValue);
                if (!toBeExpanded && !childLIElement.classList.contains('filter-hidden')) {
                    toBeExpanded = true;
                    liElement.classList.add('filter-expanded');
                }
            }
        }
        // Check if text matches
        var displayedElement = liElement.querySelector('span, a');
        var displayedText = displayedElement.textContent;
        var matches = this._stringService.contains(displayedText, filterValue);
        if (matches) {
            liElement.classList.add('filter-match');
        }
        else if (!toBeExpanded) {
            // Does not match and has no children that match
            liElement.classList.add('filter-hidden');
        }
    };
    CollapsibleMenu.prototype.restorePreFilterState = function () {
        if (this._rootElement.classList.contains('filtered')) {
            for (var i = 0; i < this._rootLIElements.length; i++) {
                this.restoreLIElementPreFilterState(this._rootLIElements[i], true);
            }
            this._rootElement.classList.remove('filtered');
        }
    };
    // Restore collapsable menu to inital state. Depth first search in post-order since child elements must be assigned their height values
    // before animations are started for parent elements (parent elements need to know how much to grow).
    CollapsibleMenu.prototype.restoreLIElementPreFilterState = function (liElement, allParentsAlreadyExpanded) {
        // Reset
        this.resetLIElementState(liElement);
        var directChildULElement = liElement.querySelector('ul');
        // Leaves don't need their heights toggled
        if (directChildULElement) {
            var alreadyExpanded = liElement.classList.contains('expanded');
            var preExpanded = liElement.classList.contains('pre-expanded');
            var closestChildLIElements = directChildULElement.children;
            liElement.classList.remove('pre-expanded');
            for (var i = 0; i < closestChildLIElements.length; i++) {
                this.restoreLIElementPreFilterState(closestChildLIElements[i], allParentsAlreadyExpanded && alreadyExpanded);
            }
            if (preExpanded && !alreadyExpanded) {
                if (allParentsAlreadyExpanded) {
                    this._heightService.toggleHeightWithTransition(directChildULElement, liElement);
                }
                else {
                    this._heightService.toggleHeightWithoutTransition(directChildULElement, liElement);
                }
            }
            else if (!preExpanded && alreadyExpanded) {
                this._heightService.toggleHeightWithTransition(directChildULElement, liElement);
            }
        }
    };
    CollapsibleMenu.prototype.savePreFilterState = function () {
        var preExpandedLIs = this.
            _rootElement.
            querySelectorAll('.expanded');
        for (var i = 0; i < preExpandedLIs.length; i++) {
            preExpandedLIs[i].classList.add('pre-expanded');
        }
    };
    CollapsibleMenu.prototype.resetLIElementState = function (liElement) {
        liElement.classList.remove('filter-hidden', 'filter-expanded', 'filter-match');
    };
    return CollapsibleMenu;
}());
exports.default = CollapsibleMenu;
//# sourceMappingURL=collapsibleMenu.js.map